"""
decodr.ciphers.transposition.railfence - rail fence (zig-zag) ciphers
"""

from __future__ import annotations

def _zigzag_indices(n: int, rails: int) -> list[int]:
    if rails < 2:
        raise ValueError("rails must be >= 2")
    rail = 0
    direction = 1
    rail_of_index: list[int] = []
    for _ in range(n):
        rail_of_index.append(rail)
        rail += direction
        if rail == rails - 1:
            direction = -1
        elif rail == 0:
            direction = 1
    
    indices: list[int] = []
    for r in range(rails):
        indices.extend(i for i, rr in enumerate(rail_of_index) if rr == r)
    return indices

def encrypt(plaintext: str, rails: int = 3) -> str:
    if rails < 2:
        raise ValueError("railst must be >= 2")
    n = len(plaintext)
    order = _zigzag_indices(n, rails)

    return "".join(plaintext[i] for i in order)

def decrypt(ciphertext: str, rails: int = 3) -> str:
    if rails < 2:
        raise ValueError("rails must be >= 2")
    n = len(ciphertext)
    order = _zigzag_indices(n, rails)

    plain = [""] * n
    for k, i in enumerate(order):
        plain[i] = ciphertext[k]

    return "".join(plain[i] for i in range(n))

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.transposition.railfence encrypt <text> [rails]\n"
        "python3 -m decodr.ciphers.transposition.railfence decrypt <text> [rails]\n"
    )
    if len(sys.argv) < 3:
        print(usage)
        sys.exit(1)

    cmd, text = sys.argv[1], sys.argv[2]
    rails = int(sys.argv[3]) if len(sys.argv) >= 4 else 3

    try:
        if cmd in("encrypt", "encode"):
            print(encrypt(text, rails))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(text, rails))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)




"""
decodr.ciphers.transposition.columnar - columnar transposition cipher
"""

from __future__ import annotations
from typing import List, Tuple

def _key_order(key: str) -> List[int]:
    indexed = list(enumerate(key))
    sorted_cols = sorted(indexed, key=lambda t: (t[1].upper(), t[0]))
    return [i for i, _ in sorted_cols]

def _chunks(seq: str, n: int) -> List[str]:
    return [seq[i: i + n] for i in range(0, len(seq), n)]

def encrypt(plaintext: str, key: str, pad: str | None = None) -> str:
    if not key:
        raise ValueError("Key cannot be empty.")
    ncols = len(key)
    rows = _chunks(plaintext, ncols)

    if pad is not None and rows and len(rows[-1]) < ncols:
        rows[-1] = rows[-1] + pad[0] * (ncols - len(rows[-1]))

    order = _key_order(key)

    out_chars: List[str] = []
    for col in order:
        for r in rows:
            if col < len(r):
                out_chars.append(r[col])
    
    return "".join(out_chars)

def decrypt(ciphertext: str, key: str, pad: str | None = None) -> str:
    if not key:
        raise ValueError("Key cannot be empty.")
    ncols = len(key)
    n = len(ciphertext)
    if ncols == 0:
        return ciphertext
    
    order = _key_order(key)
    nrows = (n + ncols - 1) // ncols

    if pad is not None:
        col_lengths = [nrows] * ncols
    else:
        full_in_last_row = n % ncols
        if full_in_last_row == 0 and n > 0:
            full_in_last_row = ncols
        
        col_lengths: list[int] = []
        for idx_in_order in range(ncols):
            orig_col = order[idx_in_order]
            length = nrows if orig_col < full_in_last_row else max(nrows - 1, 0)
            col_lengths.append(length)
            
    cols_data: list[str] = []
    pos = 0
    for length in col_lengths:
        cols_data.append(ciphertext[pos: pos + length])
        pos += length
    
    col_map = {
        order[i]: cols_data[i] for i in range(ncols)
    }

    pointers = {
        c: 0 for c in range(ncols)
    }
    rows: list[list[str]] = [[""] * ncols for _ in range(nrows)]
    for r in range(nrows):
        for c in range (ncols):
            col_str = col_map.get(c, "")
            p = pointers[c]
            if p < len(col_str):
                rows[r][c] = col_str[p]
                pointers[c] = p + 1
            else:
                rows[r][c] = ""


    plaintext = "".join(ch for row in rows for ch in row if ch != "")
    return plaintext
    
encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.transposition.columnar encrypt <text> <key> [pad]\n"
        "python3 -m decodr.transposition.columnar decrypt <text> <key> [pad]\n"
        "Notes:\n"
        "- If you used padding during encrypt, pass the same pad\n"
    )
    
    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)

    cmd, text, key = sys.argv[1:4]
    pad = sys.argv[4] if len(sys.argv) >= 5 else None

    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(text, key, pad=pad))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(text, key, pad=pad))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


"""decodr - core package."""

from importlib.metadata import version, PackageNotFoundError

__all__ = [
    "__version__",
    "__description__",
    "encode",
    "decode",
    "detect"
]

__description__ = "decodr - a modular CTF/crypto CLI toolkit for encodings, classic ciphers, and autodetection."

try:
    __version__ = version("decodr")
except PackageNotFoundError:
    __version__ = "0.1.0"

try:
    from .api import encode, decode, detect
except ImportError as e:
    raise ImportError(
        "decodr.api not found. Make sure 'src/decodr/api.py' exists."
    ) from e


"""
decodr.ciphers.modern.aes - AES, CBC, PKCS7, with IV embedded
"""

from __future__ import annotations
import base64
from typing import Union
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad, unpad

BLOCK_SIZE = 16

def _parse_key(key: Union[str, bytes], *, encoding: str = 'utf-8') -> bytes:
    if isinstance(key, bytes):
        k = key
    else:
        s = key
        if s.startswith("hex:"):
            try:
                k = bytes.fromhex(s[4:])
            except ValueError as e:
                raise ValueError("Invalid hex key after 'hex:' prefix")
        else:
            k = s.encode(encoding)
        
    if len(k) not in (16, 24, 32):
        raise ValueError("AES key must be 16, 24, or 32 bytes")
    return k

def encrypt(plaintext: str, key: Union[str, bytes], *, encoding: str = 'utf-8') -> str:
    k = _parse_key(key, encoding=encoding)
    iv = get_random_bytes(BLOCK_SIZE)
    cipher = AES.new(k, AES.MODE_CBC, iv=iv)
    ct = cipher.encrypt(pad(plaintext.encode(encoding), BLOCK_SIZE))
    blob = iv + ct
    return base64.b64encode(blob).decode('ascii')

def decrypt(ciphertext_b64: str, key: Union[str, bytes], *, encoding: str = 'utf-8') -> str:
    k = _parse_key(key, encoding=encoding)
    try:
        blob = base64.b64decode(ciphertext_b64, validate=True)
    except Exception as e:
        raise ValueError("Ciphertext must be valid base64")
    
    if len(blob) < BLOCK_SIZE or (len(blob) - BLOCK_SIZE) % BLOCK_SIZE != 0:
        raise ValueError("invalid IV + ciphertext length")
    
    iv, ct = blob[:BLOCK_SIZE], blob[BLOCK_SIZE:]
    cipher = AES.new(k, AES.MODE_CBC, iv=iv)
    pt = unpad(cipher.decrypt(ct), BLOCK_SIZE)
    return pt.decode(encoding)

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.modern.aes encrypt <text> <key>\n"
        "python3 -m decodr.ciphers.moder.aes decrypt <base64_iv_plus_ct> <key>\n"
    )

    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)

    cmd, arg1, key = sys.argv[1], sys.argv[2], sys.argv[3]
    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(arg1, key))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(arg1, key))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
            sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


"""
decodr.encodings.base32_mod - Base32 encode/decode helpers.

Features:
- UTF-8 text in/out (internal bytes conversion)
- based on https://datatracker.ietf.org/doc/html/rfc4648
- do not worry about missing padding on decode
"""

from __future__ import annotations
import base64


def _to_bytes(s: str) -> bytes:
    return s.encode('utf-8')

def _to_str(b: bytes) -> str:
    return b.decode('utf-8', errors="ignore")

def _fix_padding(s: str) -> str:
    r = len(s) % 8
    return s if r == 0 else s + ('=' * (8 - r))

def encode(text: str, *, padding: bool = True) -> str:
    raw = _to_bytes(text)
    out = base64.b32encode(raw).decode("ascii")
    return out if padding else out.rstrip("=")

def decode(b64text: str, *, strict: bool = False) -> str:
    data = b64text.strip().replace("\n", "").replace("\r", "").replace(" ", "")
    if not strict:
        data = _fix_padding(data)

    try: 
        decoded = base64.b32decode(data, casefold=True)
    except Exception as e:
        raise ValueError(f"Invalid base34 input: {e}") from e
    
    return _to_str(decoded)


if __name__ == "__main__":
    import sys

    if len(sys.argv) < 3:
        print("Usage: python3 -m decodr.encodings.base32_mod <encode|decode> <text>")
        sys.exit(1)

    cmd, text = sys.argv[1], sys.argv[2]

    try:
        if cmd == 'encode':
            print(encode(text))
        elif cmd == "decode":
            print(decode(text))
        else:
            print("Error: Unknown command. Use 'encode' or 'decode'.")
    except Exception as e:
        print(f"Error: {e}")


"""
decodr.ciphers.stream.xor - single-byte xor cipher
"""

from __future__ import annotations
from typing import Union

def _parse_key_to_int(key: Union[int, str]) -> int:
    if isinstance(key, int):
        k = key
    else:
        s = key.strip().lower()
        k = int(s, 16) if s.startswith("0x") else int(s, 10)
    if not (0 <= k <= 255):
        raise ValueError("key must be from 0 to 255")
    return k

def _xor_single(data: bytes, k: int) -> bytes:
    return bytes(b ^ k for b in data)

def encrypt(plaintext: str, key: Union[int, str], *, encoding: str = "utf-8") -> str:
    k = _parse_key_to_int(key)
    pt = plaintext.encode(encoding)
    ct = _xor_single(pt, k)
    return ct.hex()

def decrypt(ciphertext_hex: str, key: Union[int, str], *, encoding: str = 'utf-8') -> str:
    k = _parse_key_to_int(key)
    try:
        ct = bytes.fromhex(ciphertext_hex)
    except ValueError as e:
        raise ValueError('ciphertext must be hex-encoded') from e
    pt = _xor_single(ct, k)
    return pt.decode(encoding, errors='strict')

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.stream.xor encrypt <text> <key> <encoding>\n"
        "python3 -m decodr.ciphers.stream.xor decrypt <hex> <key> <encoding>\n"
    )

    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)

    cmd, arg_text, key = sys.argv[1], sys.argv[2], sys.argv[3]
    enc = sys.argv[4] if len(sys.argv) >= 5 else 'utf-8'
    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(arg_text, key, encoding=enc))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(arg_text, key, encoding=enc))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
            sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


"""
decodr.cli - command-line interface for decodr.

Provides the commands:
- encode
- decode
- detect
"""

from __future__ import annotations
import typer
from rich.console import Console
from rich.table import Table
from decodr import encode as api_encode, decode as api_decode, detect as api_detect

app = typer.Typer(
    help="decodr - decodr - a modular CTF/crypto CLI toolkit for encodings, classic ciphers, and autodetection."
)
console = Console()

@app.command()
def encode(
    scheme: str = typer.Argument(..., help="Encoding or cipher scheme (e.g., base64, caesar, rot13)"),
    text: str = typer.Argument(..., help="Text to encode"),
):
    try:
        result = api_encode(scheme, text)
        console.print(f"[bold green]Result:[/bold green] {result}")
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")


@app.command()
def decode(
    scheme: str = typer.Argument(..., help="Decoding or cipher scheme (e.g., base64, caesar, rot13)"),
    text: str = typer.Argument(..., help="Text to decode"),
):
    try:
        result = api_decode(scheme, text)
        console.print(f"[bold green]Result:[/bold green] {result}")
    except Exception as e:
        console.print(f"[bold red]Error:[/bold red] {e}")

@app.command()
def detect(
    text: str = typer.Argument(..., help="Text to automatically analyze and decode"),
    limit: int = typer.Option(5, help="Number of best results to show"),
): 
    try:
        results = api_detect(text, limit=limit)
        if not results:
            console.print("[yello]No likely encoding found.[/yellow]")
            raise typer.Exit()
        
        table = Table(title="Detection results")
        table.add_column("Rank", justify="right", style="cyan", no_wrap=True)
        table.add_column("Scheme", style="magenta")
        table.add_column("Score", justify="right", style="yellow")
        table.add_column("Result", style="white")

        for i, r in enumerate(results, 1):
            table.add_row(str(i), r["scheme"], r["score"], r["result"][:80])
        
        console.print(table)
    except Exception as e:
        console.print(f"[bold red]Error: [/bold red] {e}")

@app.callback()
def main():
    pass

if __name__ == "__main__":
    app()



"""
decodr.ciphers.classical.rot13 - ROT13 cipher via caesar(where shift is 13)
"""

from __future__ import annotations
from decodr.ciphers.classical import caesar

def encode(text: str) -> str:
    return caesar.encrypt(text, shift=13)

def decode(text: str) -> str:
    return caesar.decrypt(text, shift=13)

encrypt = encode
decrypt = decode

if __name__ == "__main__":
    import sys
    if len(sys.argv) < 3:
        print("Usage: python3 -m decodr.ciphers.classical.rot13 <encode|decode> <text>")
        sys.exit(1)
    
    cmd, text = sys.argv[1], sys.argv[2]

    try:
        if cmd in ("encode", "decode", "encrypt", "decrypt"):
            if cmd in ("encode", "encrypt"):
                print(encode(text))
            else:
                print(decode(text))
        else:
            print("Error: Unknown command. Use 'encode' or 'decode'.")
    except Exception as e:
        print(f"Error: {e}")


from decodr.detectors import autodetect
from decodr.detectors import file_magic

def test_autodetect():
    assert autodetect.detect_type("48656c6c6f")['type'] == 'hex'
    assert autodetect.detect_type("SGVsbG8=")['type'] == 'base64'
    assert autodetect.detect_type("%48%65%6C%6C%6F")['type'] == "url-encoded"
    assert autodetect.detect_type("HELLO")["type"] in ("plaintext", "unknown")

def test_file_magic(tmp_path):
    p = tmp_path / "sample.png"
    p.write_bytes(b"\x89PNG\r\n\x1a\n" + b"readadadakdjhakdhaskjd")
    assert file_magic.detect_file(str(p)) == "PNG image"




"""
decodr.ciphers.polyalphabetic.vigenere - Vignere cipher

E_i = (P_i + K_i) mod 26
D_i = (C_i - K_i) mod 26
"""

from __future__ import annotations

def _shift(ch: str, k: int, decrypt: bool = False) -> str:
    if not ch.isalpha():
        return ch
    base = ord("A") if ch.isupper() else ord("a")
    offset = -k if decrypt else k
    return chr((ord(ch) - base + offset) % 26 + base)

def _key_shifts(key: str) -> list[int]:
    return [(ord(c.lower()) - ord("a")) for c in key if c.isalpha()]

def encrypt(plaintext: str, key: str) -> str:
    if not key:
        raise ValueError('Key cannot be empty')
    shifts = _key_shifts(key)
    out = []
    j = 0
    for ch in plaintext:
        if ch.isalpha():
            out.append(_shift(ch, shifts[j % len(shifts)]))
            j += 1
        else:
            out.append(ch)
    return "".join(out)

def decrypt(ciphertext: str, key: str) -> str:
    if not key:
        raise ValueError('Key cannot be empty')
    shifts = _key_shifts(key)
    out = []
    j = 0
    for ch in ciphertext:
        if ch.isalpha():
            out.append(_shift(ch, shifts[j % len(shifts)], decrypt=True))
            j += 1
        else:
            out.append(ch)
    return "".join(out)

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.polyalphabetic.vigenere encrypt <text> <key>\n"
        "python3 -m decodr.ciphers.polyalphabetic.vigenere decrypt <text> <key>\n"
    )

    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)
    
    cmd, text, key = sys.argv[1:4]

    try:
        if cmd in("encrypt", "encode"):
            print(encrypt(text, key))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(text, key))
        else:
            print("Unknown command. Use 'decrypt' or 'encrypt'.")
            print(usage)
    except Exception as e:
        print(f"Error {e}")
        sys.exit(1)




"""
decodr.ciphers.modern.hashes = common has algoritms md5, sha1, sha256/512
"""

from __future__ import annotations
import hashlib
from typing import Literal
import hmac

HashAlgo = Literal[
    "md5",
    "sha1",
    "sha256",
    "sha512"
]

def hash_text(text: str, algo: HashAlgo = "sha256", *, encoding: str = "utf-8") -> str:
    algo = algo.lower()
    if algo not in hashlib.algorithms_available:
        raise ValueError(f"Unsupported has algoritm: {algo}")
    h = hashlib.new(algo)
    h.update(text.encode(encoding))
    return h.hexdigest()

def verify_hash(text: str, digest: str, algo: HashAlgo = "sha256", *, encoding: str = "utf-8") -> bool:
    return hmac.compare_digest((hash_text(text, algo, encoding=encoding)), digest)

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.modern.hashes hash <text> [algo]\n"
        "python3 -m decodr.ciphers.modern.hashes verify <text> <digest>"
    )

    if len(sys.argv) < 3 :
        print(usage)
        sys.exit(1)

    cmd = sys.argv[1]
    try:
        if cmd == "hash":
            text = sys.argv[2]
            algo = sys.argv[3] if len(sys.argv) >= 4 else "sha256"
            print(hash_text(text, algo))
        elif cmd == "verify":
            if len(sys.argv) < 4:
                print("Usage: verify <text> <digest> [algo]")
                sys.exit(1)
            text, digest = sys.argv[2], sys.argv[3]
            algo = sys.argv[4] if len(sys.argv) >= 5 else "sha256"
            ok = verify_hash(text, digest, algo)
            print("✅ match" if ok else "❌ not a match")
        else:
            print(usage)
            sys.exit(1)

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


"""
decodr.encodings.hex_mod - Hexadecimal encode/decode helpers

Supports:
- encoding UTF-8 text to hex
- decoding hex strings 
"""

from __future__ import annotations
import binascii

def _to_bytes(s: str) -> bytes:
    return s.encode('utf-8')

def _to_str(b: bytes) -> str:
    return b.decode('utf-8', errors="ignore")

def encode(text: str, uppercase: bool = False, spaced: bool = False) -> str:
    raw = _to_bytes(text)
    hex = binascii.hexlify(raw).decode('ascii')
    if uppercase:
        hex = hex.upper()
    if spaced:
        hex = " ".join(hex[i:i+2] for i in range (0, len(hex), 2)) 
    
    return hex

def decode(hextext: str, strict: bool = True) -> str:
    data = hextext.strip().replace(" ", "").replace("\n", "").replace("\r", "")

    try:
        decoded = binascii.unhexlify(data)
    except Exception as e:
        if strict:
            raise ValueError(f"Error: Invalid hex input: {e}") from e
        else:
            cleaned = "".join(ch for ch in data if ch.lower() in "0123456789abcdef")
            decoded = binascii.unhexlify(cleaned)
    return _to_str(decoded)


if __name__ == "__main__":
    import sys

    if len(sys.argv) < 3:
        print("Usage: python3 -m decodr.encodings.hex_mod <encode|decode> <text>")
        sys.exit(1)

    cmd, text = sys.argv[1], sys.argv[2]

    try:
        if cmd == 'encode':
            print(encode(text))
        elif cmd == "decode":
            print(decode(text))
        else:
            print("Error: Unknown command. Use 'encode' or 'decode'.")
    except Exception as e:
        print(f"Error: {e}")


"""
decodr.encodings.url_mod - URL encoding helpers.

This module just implements encode/decode for URL-safe percent-encoding.
https://datatracker.ietf.org/doc/html/rfc3986
"""

from __future__ import annotations
from urllib.parse import quote, unquote

def encode(text: str, safe: str = '/') -> str:
    return quote(text, safe=safe)

def decode(text: str) -> str:
    return unquote(text)

if __name__ == "__main__":
    import sys

    if len(sys.argv) < 3:
        print("Usage: python3 -m decodr.encodings.url_mode <encode|decode> <text>")
        sys.exit(1)

    cmd, text = sys.argv[1], sys.argv[2]

    try:
        if cmd == 'encode':
            print(encode(text))
        elif cmd == "decode":
            print(decode(text))
        else:
            print("Error: Unknown command. Use 'encode' or 'decode'.")
    except Exception as e:
        print(f"Error: {e}")




"""
decodr.ciphers.rotor.enigma - its a simplied version for my sake i am not alan turing
"""

from __future__ import annotations
from typing import List, Dict

# to do for the the next version: let the user choose its own rotors

ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

ROTOR_I = "EKMFLGDQVZNTOWYHXUSPAIBRCJ"
ROTOR_II = "AJDKSIRUXBLHWTMCQGZNPYFVOE"
ROTOR_III = "BDFHJLCPRTXVZNYEIWGAKMUSQO"
REFLECTOR_B = "YRUHQSLDPXNGOKMIEBFZCWVJAT"

def _shift_rotor(rotor: str, shift: int) -> str:
    s = shift % 26
    return rotor[s:] + rotor[:s]

def _enc_through_rotor(ch: str, rotor: str) -> str:
    return rotor[ALPHABET.index(ch)]

def _dec_through_rotor(ch: str, rotor: str) -> str:
    return ALPHABET[rotor.index(ch)]

def encrypt(plaintext: str) -> str:
    r1, r2, r3 = ROTOR_I, ROTOR_II, ROTOR_III 
    s1 = s2 = s3 = 0
    out: List[str] = []

    for ch in plaintext.upper():
        if ch not in ALPHABET:
            out.append(ch)
            continue

        s1 = (s1 + 1) % 26
        if s1 == 0:
            s2 = (s2 + 1) % 26
            if s2 == 0:
                s3 = (s3 + 1) % 26
        
        a = _enc_through_rotor(ch, _shift_rotor(r1, s1))
        b = _enc_through_rotor(a, _shift_rotor(r2, s2))
        c = _enc_through_rotor(b, _shift_rotor(r3, s3))

        r = REFLECTOR_B[ALPHABET.index(c)]

        c2 = _dec_through_rotor(r, _shift_rotor(r3, s3))
        b2 = _dec_through_rotor(c2, _shift_rotor(r2, s2))
        a2 = _dec_through_rotor(b2, _shift_rotor(r1, s1))

        out.append(a2)

    return "".join(out)

decrypt = encrypt
encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.rotor.enigma encrypt <text>\n"
        "python3 -m decodr.ciphers.rotor.enigma decrypt <text>\n"
    )

    if len(sys.argv) < 3:
        print(usage)
        sys.exit(1)
    
    cmd, text = sys.argv[1], sys.argv[2]

    try:
        if cmd in ("encrypt", "encode", "decrypt", "decode"):
            print(encrypt(text))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
            sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)
        


"""
decodr.encodings - collection of encoding/decoding helpers.

It includes:
- base64_mod
- base32_mod
- hex_mod
- url_mod
"""

from __future__ import annotations

__all__ = [
    "base64_mod", 
    "base32_mod",
    "hex_mod",
    "url_mod"
]


"""
decodr.ciphers.classical.caesar - Caesar cipher - a monoalphabetic shift

Features:
- encrypt or decript using a integer shift
- non-letters are gonna be preserved as is
"""

from __future__ import annotations

def _shift_char(ch: str, shift: int) -> str:
    if "a" <= ch <= "z":
        base = ord("a")
        return chr((ord(ch) - base + shift) % 26 + base)
    if "A" <= ch <= "Z":
        base = ord("A")
        return chr((ord(ch) - base + shift) % 26 + base)
    return ch

def encrypt(plaintext: str, shift: int = 3) -> str:
    s = shift % 26
    return "".join(_shift_char(ch, s) for ch in plaintext)

def decrypt(ciphertext: str, shift: int = 3) -> str:
    s = (-shift)%26
    return "".join(_shift_char(ch, s) for ch in ciphertext)

def encode(text: str, shift: int = 3) -> str:
    return encrypt(text, shift=shift)

def decode(text: str, shift: int = 3) -> str:
    return decrypt(text, shift=shift)
    
def crack(ciphertext: str) -> list[tuple[int, str]]:
    results: list[tuple[int, str]] = []
    for sh in range(26):
        pt = decrypt(ciphertext, shift=sh)
        results.append((sh, pt))
    return results

if __name__ == "__main__":
    import sys
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.classical.caesar encrypt <text> [shift]\n"
        "python3 -m decodr.ciphers.classical.caesar decrypt <text> [shift]\n"
        "python3 -m decodr.ciphers.classical.caesar crack <text>\n"
    )
    if(len(sys.argv) < 3):
        print(usage)
        sys.exit(1)
    
    cmd = sys.argv[1]
    text = sys.argv[2]
    shift = int(sys.argv[3]) if len(sys.argv) >= 4 else 3

    try:
        if cmd == "encrypt":
            print(encrypt(text, shift=shift))
        elif cmd == "decrypt":
            print(decrypt(text, shift=shift))
        elif cmd == "crack":
            for sh, pt in crack(text):
                print(f"{sh:2d}: {pt}")
        else:
            print("Error: Unknown command. Use 'encrypt', 'decrypt' or 'crack'.")
            print(usage)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)




"""
decodr.ciphers.classical.affine - affine cipher 

Encryption: E(x) = (ax + b) mod 26
Decryption: D(x) = a_inv * (x - b) mod 26
a is the modulare inverse of a mod 26
"""

from __future__ import annotations

def _char_to_num(ch: str) -> int:
    return ord(ch.lower()) - ord('a')

def _num_to_char(n: int, upper: bool) -> str:
    base = ord("A") if upper else ord("a")
    return chr((n % 26) + base)

def _modinv(a: int, m: int) -> int:
    a = a % m
    for x in range(1, m):
        if (a * x) % m == 1:
            return x
    raise ValueError(f"There is no modular inverse for a={a} mod {m}")

def encrypt(plaintext: str, a: int = 5, b: int = 8) -> str:
    result = []
    for ch in plaintext:
        if ch.isalpha() and ch.isascii():
            upper = ch.isupper()
            x = _char_to_num(ch)
            y = (a * x + b) % 26
            result.append(_num_to_char(y, upper))
        else:
            result.append(ch)
    return "".join(result)

def decrypt(ciphertext: str, a: int = 5, b: int = 8) -> str:
    result = []
    a_inv = _modinv(a, 26)
    for ch in ciphertext:
        if ch.isalpha() and ch.isascii():
            upper = ch.isupper()
            y = _char_to_num(ch)
            x = (a_inv * (y - b) % 26)
            result.append(_num_to_char(x, upper))
        else:
            result.append(ch)
    return "".join(result)

encode = encrypt
decode = decrypt

def crack(ciphertex: str) -> list[tuple[int, int, str]]:
    valid_a = [1,3,5,7,9,11,15,17,19,21,23,25]
    results: list[tuple[int, int, str]] = []
    for a in valid_a:
        for b in range(26):
            try:
                pt = decrypt(ciphertex, a, b)
                results.append((a, b, pt))
            except Exception:
                continue
    return results

if __name__ == "__main__":
    import sys 

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.classical.affine encrypt <text> [a] [b]\n"
        "python3 -m decodr.ciphers.classical.affine decrypt <text> [a] [b]\n"
        "python3 -m decodr.ciphers.classical.affine crack <ciphertext>\n"
    )

    if(len(sys.argv) < 3):
        print(usage)
        sys.exit(1)

    cmd, text = sys.argv[1], sys.argv[2]
    try:
        a = int(sys.argv[3]) if len(sys.argv) >= 4 else 5
        b = int(sys.argv[4]) if len(sys.argv) >= 4 else 8
    except ValueError:
        print('a and b must be integers')
        sys.exit(1)

    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(text, a, b))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(text, a, b))
        elif cmd == "crack":
            for a_val, b_val, pt in crack(text):
                print(f"a={a_val:2d}, b={b_val:2d} -> {pt}")
        else:
            print("Error: Unknown command. Use 'encrypt', 'decrypt' or 'crack'.")
            print(usage)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)

    





"""
decodr.ciphers.modern.rsa - rsa cipher
"""

from __future__ import annotations
from typing import Tuple
import math

def _gcd(a: int, b: int) -> int:
    while b:
        a, b = b, a % b
    return a

def _modinv(a: int, m: int) -> int:
    r0, r1 = a, m
    s0, s1 = 1, 0
    while r1 != 0:
        q = r0 // r1
        r0, r1 = r1, r0 - q * r1
        s0, s1 = s1, s0 - q * s1
    
    if r0 != 1:
        raise ValueError('Modular inverse does not exist.')

    return s0 % m

def is_prime(n: int) -> bool:
    if n <= 1:
        return False
    if n <= 3:
        return True
    if n % 2 == 0 or n % 3 == 0:
        return False
    i = 5
    while i * i <= n:
        if n % i == 0 or n % (i + 2) == 0:
            return False
        i += 6

    return True

def _generate_keys(p: int, q: int, e: int = 65537) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    if p == q:
        raise ValueError("p and q must be distinct primes")
    if not (is_prime(p) and is_prime(q)):
        raise ValueError("p and q must both be primes")
    n = p * q
    phi = (p - 1) * (q - 1)
    if _gcd(e, phi) != 1:
        raise ValueError("e must be coprime with phi(n)")
    
    d = _modinv(e, phi)
    return (n, e), (n, d)

def encrypt(plaintext: str, n: int, e: int) -> str:
    return " ".join(str(pow(ord(ch), e, n)) for ch in plaintext)

def decrypt(ciphertext: str, n: int, d: int) -> str:
    parts = ciphertext.strip().split()
    return "".join(chr(pow(int(x), d, n)) for x in parts)

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys

    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.modern.rsa gen <p> <q> [e]\n"
        "python3 -m decodr.ciphers.modern.rsa encrypt <text> <n> <e>\n"
        "python3 -m decodr.ciphers.modern.rsa decrypt <cipher> <n> <d>\n"
    )

    if len(sys.argv) < 3:
        print(usage)
        sys.exit(1)

    cmd = sys.argv[1]
    try:
        if cmd == "gen":
            if len(sys.argv) < 3:
                print("Usage <p> <q> [e]")
                sys.exit(1)
            p, q = int(sys.argv[2]), int(sys.argv[3])
            e = int(sys.argv[4]) if len(sys.argv) >= 5 else 65537
            pub, priv = _generate_keys(p, q, e)
            print(f"Public key (n, e): {pub}")
            print(f"Private key (n, d): {priv}")

        elif cmd in ("encrypt", "encode"):
            if len(sys.argv) < 5:
                print("Usage: encrypt <text> <n> <e>")
                sys.exit(1)
            text, n, e = sys.argv[2], int(sys.argv[3]), int(sys.argv[4])
            print(encrypt(text, n, e))
        
        elif cmd in ("decrypt", "decode"):
            if len(sys.argv) < 5:
                print("Usage: decrypt <cipher> <n> <d>")
                sys.exit(1)
            cipher, n, d = sys.argv[2], int(sys.argv[3]), int(sys.argv[4])
            print(decrypt(cipher, n, d))

        else:
            print(usage)
            sys.exit(1)

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)





"""
decodr.ciphers.stream.rc4 - RC4 stream cipher
"""

from __future__ import annotations
from typing import Union, List, Generator

def _ksa(key_bytes: bytes) -> List[int]:
    key_length = len(key_bytes)
    S = list(range(256))
    j = 0
    for i in range(256):
        j = (j + S[i] + key_bytes[i % key_length]) % 256
        S[i], S[j] = S[j], S[i]

    return S

def _prga(S: List[int]) -> Generator[int, None, None]:
    i, j = 0, 0
    while True:
        i = (i + 1) % 256
        j = (j + S[i]) % 256
        S[i], S[j] = S[j], S[i]
        K = S[(S[i] + S[j]) % 256]
        yield K

def _rc4_encrypt(data: bytes, key_bytes: bytes) -> bytes:
    S = _ksa(key_bytes)
    keystream = _prga(S)
    return bytes(b ^ next(keystream) for b in data)

def encrypt(plaintext: str, key: Union[str, bytes], *, encoding: str = 'utf-8') -> str:
    key_b = key.encode(encoding) if isinstance(key, str) else key
    if not key_b:
        raise ValueError("Key cannot be empty.")

    pt = plaintext.encode(encoding)
    ct = _rc4_encrypt(pt, key_b)
    return ct.hex()

def decrypt(ciphertext_hex: str, key: Union[str, bytes], *, encoding: str = 'utf-8') -> str:
    key_b = key.encode(encoding) if isinstance(key, str) else key
    if not key_b:
        raise ValueError("Key cannot be empty.")
    try:
        ct = bytes.fromhex(ciphertext_hex)
    except ValueError as e:
        raise ValueError("Ciphertext must be hex-encoded") from e
    pt = _rc4_encrypt(ct, key_b)
    return pt.decode(encoding, errors='strict')

encode = encrypt
decode = decrypt


if __name__ == "__main__":
    import sys
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.stream.rc4 encrypt <text> <key> <encoding>\n"
        "python3 -m decodr.ciphers.stream.rc4 decrypt <hex>  <key> <encoding>\n"
    )

    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)

    cmd, arg_text, key = sys.argv[1], sys.argv[2], sys.argv[3]
    enc = sys.argv[4] if len(sys.argv) >= 5 else 'utf-8'
    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(arg_text, key, encoding=enc))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(arg_text, key, encoding=enc))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
            sys.exit(1)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


"""
decodr.ciphers.classical.substitution - simple monoalphabetic substution
"""

from __future__ import annotations
import sys
import string
import random
from typing import Tuple, Dict, List

ALPHABET_LO = string.ascii_lowercase
ALPHABET_UP = string.ascii_uppercase

def normalize_key(key: str) -> str:
    return "".join(ch for ch in key if ch.isalpha()).upper()

def validate_key(key: str) -> bool:
    k = normalize_key(key)
    return len(k) == 26 and set(k) == set(ALPHABET_UP)

def random_key() -> str:
    letters = list(ALPHABET_UP)
    random.shuffle(letters)
    return "".join(letters)

def _build_maps(key: str) -> Tuple[Dict[str, str], Dict[str, str]]:
    k = normalize_key(key)
    forward = {ALPHABET_LO[i]: k[i].lower() for i in range(26)}
    forward.update({ALPHABET_UP[i]: k[i] for i in range(26)})
    reverse = { v: u for u, v in forward.items() }
    return forward, reverse

def encrypt(plaintext: str, key: str) -> str:
    if not validate_key(key):
        raise ValueError("Invalid key: must be 26 unique letters A-Z")
    forward, _ = _build_maps(key)
    return "".join(forward.get(ch, ch) for ch in plaintext)

def decrypt(plaintext: str, key: str) -> str:
    if not validate_key(key):
        raise ValueError("Invalid key: must be 26 unique letters A-Z")
    _, reverse = _build_maps(key)
    return "".join(reverse.get(ch, ch) for ch in plaintext)

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.classical.substitution generate-key\n"
        "python3 -m decodr.ciphers.classical.substitution encrypt <text> <key>\n"
        "python3 -m decodr.ciphers.classical.substitution decrypt <text> <key>\n"
    )

    if len(sys.argv) < 2:
        print(usage)
        sys.exit(1)
    
    cmd = sys.argv[1]
    try:
        if cmd == "generate-key":
            print(random_key())
        elif cmd in ("encode", "encrypt"):
            if len(sys.argv) < 4:
                print("encrypt requires <text> <key>")
                sys.exit(1)
            text, key = sys.argv[2], sys.argv[3]
            print(encrypt(text, key))
        elif cmd in ("decrypt", "decode"):
            if len(sys.argv) < 4:
                print("encrypt requires <text> <key>")
                sys.exit(1)
            text, key = sys.argv[2], sys.argv[3]
            print(decrypt(text, key))
        else:
            print("Unknown command.")
            print(usage)
            sys.exit(1)
    except Exception as e:
        print("Error:", e)
        sys.exit(1)
        


"""
decodr.utils.ioutils - input/output helpers.

Provides helper functions for reading input and ofc writing output,
that's it either from strings, files, or stdin/stdout.
"""

from __future__ import annotations
from pathlib import Path
import sys

def read_input(text: str | None = None, infile: str | None = None, binary: bool = False) -> bytes:
    if text is not None:
        return text.encode('utf-8') if not binary else text.encode('utf-8', 'ignore')
    
    if infile:
        path = Path(infile)
        # mode = "rb" if binary else "r"
        data = path.read_bytes() if binary else path.read_text(encoding="utf-8", errors="ignore")
        return data if isinstance(data, bytes) else data.encode('utf-8')
    
    data = sys.stdin.buffer.read()
    return data if isinstance(data, bytes) else data.encode('utf-8', 'ignore') 

def write_output(data: bytes, outFile: str | None = None, binary: bool = False):
    if outFile:
        path = Path(outFile)
        if binary:
            path.write_bytes(data)
        else:
            path.write_text(data.decode('utf-8', 'ignore'), encoding='utf-8')
    else:
        if binary:
            sys.stdout.buffer.write(data)
        else:
            print(data.decode('utf-8', 'ignore'))


"""
decodr.ciphers.classical.atbash - Atbash cipher

Basically each letter is mapped to its reverse counterpard:
A <-> Z, B <-> Y ...
"""

from __future__ import annotations

def _map_char(ch: str) -> str:
    if "a" <= ch <= "z":
        return chr(ord("z") - (ord(ch) - ord("a")))
    if "A" <= ch <= "Z":
        return chr(ord("Z") - (ord(ch) - ord("A")))
    return ch

def transform(text: str) -> str:
    return "".join(_map_char(ch) for ch in text)

def encode(text: str) -> str:
    return transform(text)

decode = encode
encrypt = encode
decrypt = decode

if __name__ == "__main__":
    import sys
    if len(sys.argv) < 3:
        print("Usage: python3 -m decodr.ciphers.classical.atbash <encode|decode> <text>")
        sys.exit(1)
    
    cmd, text = sys.argv[1], sys.argv[2]
    try:
        if cmd in ("encode", "decode", "encrypt", "decrypt"):
            print(transform(text))
        else:
            print("Error: Unknown command. Use 'encode' or 'decode'.")
    except Exception as e:
        print(f"Error: {e}")




"""
decodrs.ciphers.polyalphabetic.playfair - playfair cipher (5x5, I/J merged)
"""

from __future__ import annotations
from typing import List, Tuple

ALPHABET_25 = "ABCDEFGHIKLMNOPQRSTUVWXYZ"

def _normalize_key(key: str) -> str:
    seen = set()
    out = []
    for ch in key.upper():
        if not ch.isalpha():
            continue
        ch = "I" if ch == "J" else ch
        if ch not in seen and ch in ALPHABET_25:
            seen.add(ch)
            out.append(ch)
        
    return "".join(out)

def _build_square(key: str) -> List[List[str]]:
    norm = _normalize_key(key)
    rest = "".join(ch for ch in ALPHABET_25 if ch not in norm)
    seq = norm + rest
    return [list(seq[i: i + 5]) for i in range(0, 25, 5)]

def _pos(square: List[List[str]], ch: str) -> Tuple[int, int]:
    for r in range(5):
        for c in range(5):
            if square[r][c] == ch:
                return r, c
    raise ValueError(f"Letter {ch!r} not found in square")

def _prepare_text(text: str, pad: str = "X") -> List[Tuple[str, str]]:
    s = []
    for ch in text.upper():
        if ch.isalpha():
            s.append("I" if ch == "J" else ch)
    
    pairs: List[Tuple[str, str]] = []
    i = 0
    while i < len(s):
        a = s[i]
        b = s[i + 1] if i + 1 < len(s) else None
        if b is None:
            pairs.append((a, pad))
            i += 1
        elif a == b:
            pairs.append((a, pad))
            i += 1
        else:
            pairs.append((a, b))
            i += 2
    return pairs

def __process_pair(a: str, b: str, square: List[List[str]], decrypt: bool = False) -> Tuple[str, str]:
    ra, ca = _pos(square, a)
    rb, cb = _pos(square, b)

    if ra == rb:
        if decrypt:
            return square[ra][(ca - 1) % 5], square[rb][(cb - 1) % 5] 
        else:
            return square[ra][(ca + 1) % 5], square[rb][(cb + 1) % 5]
    
    if ca == cb:
        if decrypt:
            return square[(ra - 1) % 5][ca], square[(rb - 1) % 5][cb]
        else:
            return square[(ra + 1) % 5][ca], square[(rb + 1) % 5][cb]
    
    return square[ra][cb], square[rb][ca]

def encrypt(plaintext: str, key: str, pad: str = "X") -> str:
    square = _build_square(key)
    digrams = _prepare_text(plaintext, pad=pad)
    out: List[str] = []
    for a, b in digrams:
        ea, eb = __process_pair(a, b, square, decrypt=False)
        out.extend([ea, eb])
    
    return "".join(out)

def decrypt(ciphertext: str, key: str, pad: str = "X") -> str:
    square = _build_square(key)
    s = [("I" if ch == "J" else ch) for ch in ciphertext.upper() if ch.isalpha()]
    if len(s) % 2 != 0:
        raise ValueError("Ciphertext length must be even!")
    out: List[str] = []
    for i in range(0, len(s), 2):
        a, b = s[i], s[i + 1]
        da, db = __process_pair(a, b, square, decrypt=True)
        out.extend([da, db])

    return "".join(out)

encode = encrypt
decode = decrypt

if __name__ == "__main__":
    import sys
    usage = (
        "Usage:\n"
        "python3 -m decodr.ciphers.polyalphabetic.playfair encrypt <text> <key>\n"
        "python3 -m decodr.ciphers.polyalphabetic.playfair decrypt <text> <key>\n"
    )
    if len(sys.argv) < 4:
        print(usage)
        sys.exit(1)

    cmd, text, key = sys.argv[1:4]
    try:
        if cmd in ("encrypt", "encode"):
            print(encrypt(text, key))
        elif cmd in ("decrypt", "decode"):
            print(decrypt(text, key))
        else:
            print("Unknown command. Use 'encrypt' or 'decrypt'.")
            print(usage)
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)